# -*- coding: utf-8 -*-
"""
Created on Sat Dec  3 16:01:44 2022

@author: MSH
"""


"""Collection of functions related to data."""

import torch
import numpy as np
import pandas as pd
from typing import Any, Callable, Dict, List, Tuple, Union
from torch.utils.data import DataLoader, Dataset
from torch.utils.data.sampler import Sampler
from torch_geometric.data import HeteroData
from .hetero_static_graph_temporal_signal import HeteroStaticGraphTemporalSignal as HGT


class MetaDataset(Dataset):
    
    """Dataset for metalearning

    Parameters
    ----------
    X : np.ndarray
        Full features dataset of shape `(n_samples,1,lookback)`, are lags  of Y

  
    B : np.ndarray
        Full base forecasts dataset of shape `(n_samples, n_forecasters, horizon)`.
        
    Y : np.ndarray
        Full targets dataset of shape `(n_samples, horizon)`.
        
    Z : np.ndarray
        Full targets dataset of shape `(n_samples, n_features, lookback + horizon)`.
        
    Feat : None or array-like
        If not None then of shape `(n_samples,n_forecasters)` representing ts features for each sample.

    Yc : np.ndarray
        Full targets dataset of shape `(n_samples, n_forecasters)`representing best model indicator for each sample.
        
    Ye : np.ndarray
        Full targets dataset of shape `(n_samples, n_forecasters)`representing losses of base_forecasters for each sample.  
    
    y : np.ndarray
        Full targets dataset of shape `(n_samples, horizon)`representing the meta forecast for each sample. 
        
    transform : None or callable
        If provided, then a callable that transforms a single sample.
    """

    def __init__(
        self, 
        X : np.ndarray, 
        B: np.ndarray = None,
        Y: np.ndarray = None, 
        Z: np.ndarray = None, 
        MetaFeat: np.ndarray = None, 
        transform: np.ndarray = None,
        Xt_idx: pd.Index = None,
        Yt_tidx: pd.Index = None,
        df_idx: pd.MultiIndex = None,
        features_num: List[str] = None,
        features_cat: List[str] = None,
        forecasters: List[str] = None,
        static_known_features: List[str] = None,
        dynamic_known_features_num : List[str] = None,
        dynamic_known_features_cat : List[str] = None,
        mask: np.ndarray = None,  
        weight: np.ndarray = None, 
        edge_index:List[Union[np.ndarray, None]] = None,
        edge_weight: List[Union[np.ndarray, None]] = None,
    ):
        
        
        """Construct."""
        super().__init__()
        self.X = X
        self.B = B
        self.Y = Y
        self.Z = Z
        
        self.MetaFeat = MetaFeat
        
        self.Xt_idx =  Xt_idx
        self.Yt_tidx =  Yt_tidx
        self.df_idx = df_idx
        self.features_num = features_num
        self.features_cat = features_cat        
        self.mask = mask  if mask is not None else np.ones( (X.shape[0], X.shape[-1] + Y.shape[-1]))
        self.weight = weight if weight is not None else np.ones( (X.shape[0], X.shape[-1] + Y.shape[-1]))
        self.forecasters = forecasters  if forecasters is not None else [] 
        self.static_known_features = static_known_features if static_known_features is not None else [] 
        self.dynamic_known_features_num = dynamic_known_features_num 
        self.dynamic_known_features_cat = dynamic_known_features_cat     
        
        self.feature_names = self.features_num + self.features_cat
        self.known_features = self.dynamic_known_features_num + self.dynamic_known_features_cat + self.static_known_features
        self.unknown_features = [ff for ff in self.feature_names if ff not in self.known_features  ]
        
        self.edge_index = edge_index
        self.edge_weight = edge_weight
        
        self.features_num_idx = [self.feature_names.index(i)-1 for i in self.features_num[1:]]
        
        self.features_cat_idx = [self.feature_names.index(i)-1 for i in self.features_cat]

        self.static_known_features_idx = [self.feature_names.index(i)-1 for i in self.static_known_features]      

        self.dynamic_known_features_num_idx = [self.feature_names.index(i)-1 for i in self.dynamic_known_features_num]  

        self.dynamic_known_features_cat_idx = [self.feature_names.index(i)-1 for i in self.dynamic_known_features_cat] 

        self.known_features_idx =  self.dynamic_known_features_num_idx + self.dynamic_known_features_cat_idx + self.static_known_features_idx
        
        self.unknown_features_idx = [self.feature_names.index(i)-1 for i in self.unknown_features]
        
        self.transform = transform
        # utility
        self.n_features, self.lookback = Z.shape[1], X.shape[2]
        
        if self.Y is not None:
            self.horizon = Y.shape[1]
            if self.B is not None:
                self.BestModel()
            else:
                self.Yc = None
                self.Ye = None
                self.err = None
                
        if self.transform:
            self.Y_trans, self.mask_trans = self.transform(self.Y, self.mask[:,self.lookback:(self.lookback+self.horizon)]  )                
            self.weight_trans = self.transform(self.weight[:,self.lookback:(self.lookback+self.horizon)]  )   
        

    def __len__(self):
        """Compute length."""
        return len(self.X)

    def __getitem__(self, ix: int) -> Dict[str, torch.Tensor]:
        """Get item."""
        if self.Y is not None:
            if self.transform:
               Y_sample = torch.from_numpy(self.Y_trans[ix]).type(torch.float32) 
               mask_sample = torch.from_numpy(self.mask_trans[ix]).type(torch.float32) 
               weight_sample = torch.from_numpy(self.weight_trans[ix]).type(torch.float32) 
            else:
               Y_sample = torch.from_numpy(self.Y[ix]).type(torch.float32)
               mask_sample = torch.from_numpy(self.mask[ix,self.lookback:(self.lookback+self.horizon)]).type(torch.float32)
               weight_sample = torch.from_numpy(self.weight[ix,self.lookback:(self.lookback+self.horizon)]).type(torch.float32)
        else:
            Y_sample = torch.zeros(1, 1)            
            mask_sample = torch.zeros(1, 1)
            weight_sample = torch.zeros(1, 1)
            
        
        X_sample = torch.from_numpy( np.concatenate([self.X[ix],self.Z[ix,:,0:self.lookback]],0) ).type(torch.float32) \
            if self.Z is not None else torch.from_numpy( self.X[ix] ).type(torch.float32)
        B_sample = torch.from_numpy(self.B[ix]).type(torch.float32)        
        E_sample = torch.from_numpy(self.err[ix]).type(torch.float32)  
       
        
        #Y_sample = torch.from_numpy(self.Y[ix]).type(torch.float32) if self.Y is not None else torch.zeros(1, 1)
        Z_sample = torch.from_numpy(self.Z[ix,self.known_features_idx,self.lookback:]).type(torch.float32)  if self.Z is not None else torch.zeros(1, 1)
        
        MetaFeat_sample = torch.from_numpy(self.MetaFeat[ix]).type(torch.float32) if self.MetaFeat is not None else torch.zeros(1, 1)          
        Yc_sample = torch.from_numpy(self.Yc[ix]).type(torch.float32) if self.Y is not None else torch.zeros(1, 1)
        Ye_sample = torch.from_numpy(self.Ye[ix]).type(torch.float32) if self.Y is not None else torch.zeros(1, 1)
        
        #mask_sample = torch.from_numpy(self.mask[ix]).type(torch.float32)         
        

        return dict(
                X = torch.nan_to_num(X_sample,nan = -1),
                B = torch.nan_to_num(B_sample,nan = 0), 
                Y = torch.nan_to_num(Y_sample,nan = 0),
                Z = torch.nan_to_num(Z_sample,nan = -1), 
                mask = torch.nan_to_num(mask_sample,0), 
                metafeat = torch.nan_to_num(MetaFeat_sample,nan =0), 
                Yc = torch.nan_to_num(Yc_sample,nan=0),
                Ye = torch.nan_to_num(Ye_sample,nan=0),
                weight = torch.nan_to_num(weight_sample,nan=0),
                E = torch.nan_to_num(E_sample,nan=0),
                )

    @staticmethod
    def _collate_fn(
        batches: List[Dict[str, torch.Tensor]]
    ) -> Dict[str, torch.Tensor]:
        """
        Collate function to combine items into mini-batch for dataloader.

        Args:
            batches (List[Tuple[Dict[str, torch.Tensor], torch.Tensor]]): List of samples generated with
                :py:meth:`~__getitem__`.

        Returns:
            Tuple[Dict[str, torch.Tensor], Tuple[Union[torch.Tensor, List[torch.Tensor]], torch.Tensor]: minibatch
        """
        # collate function for dataloader

        X = torch.stack([batch["X"] for batch in batches])
        B = torch.stack([batch["B"] for batch in batches])
        Y = torch.stack([batch["Y"] for batch in batches])
        Z = torch.stack([batch["Z"] for batch in batches])
        mask = torch.stack([batch["mask"] for batch in batches])
        metafeat = torch.stack([batch["metafeat"] for batch in batches])
        Yc = torch.stack([batch["Yc"] for batch in batches])
        Ye = torch.stack([batch["Ye"] for batch in batches])
        weight = torch.stack([batch["weight"] for batch in batches])       
        E = torch.stack([batch["E"] for batch in batches]) 

        return dict(
                X = X,
                B = B,
                Y = Y,
                Z = Z,
                mask = mask,
                metafeat = metafeat,
                Yc = Yc,
                Ye = Ye,
                weight = weight,
                E = E,
                )
        
    def to_dataloader(
        self, train: bool = True, batch_size: int = 1024, batch_sampler: Union[Sampler, str] = None, **kwargs
    ) -> DataLoader:
        
        """
        Get dataloader from dataset.
        """
        default_kwargs = dict(
            shuffle=train,
            drop_last=train and len(self) > batch_size,
            collate_fn=self._collate_fn,
            batch_size=batch_size,
            batch_sampler=batch_sampler,
        )
        default_kwargs.update(kwargs)
        kwargs = default_kwargs

        return DataLoader(
            self,
            **kwargs,
        )        
    
    def BestModel(self):
        B = np.nan_to_num(self.B , nan  = 0)
      
        Y, mask  = self.Y, self.mask[:,self.lookback:(self.lookback+self.horizon)]
        
        if self.transform:
           Y, mask = self.transform(self.Y,self.mask[:,self.lookback:(self.lookback+self.horizon)])
        
        while len(self.B.shape) > len(Y.shape) :    
            Y = np.expand_dims(np.nan_to_num(Y,nan = 0), axis =1) 
            mask = np.expand_dims(np.nan_to_num(mask,nan = 0), axis =1) 
        err = (Y-B)*mask
        loss = np.nanmean((err)**2,-1)
        if len(loss.shape) ==3:
            loss = loss.mean(-1)
        Yc = loss == loss.min(axis=1,keepdims = True)
        self.Yc, self.Ye, self.err = Yc, loss, (Y-B)
        
    def evaluate(self,pred,metrics):
        ## further consider mask
        Y, mask  = self.Y, self.mask[:,self.lookback:(self.lookback+self.horizon)]
        
        if self.transform:
           Y, mask = self.transform(self.Y,self.mask[:,self.lookback:(self.lookback+self.horizon)])
       
        while len(self.B.shape) > len(Y.shape) :
            Y = np.expand_dims(np.nan_to_num(Y,nan = 0), axis =1) 
            mask = np.expand_dims(np.nan_to_num(mask,nan = 0), axis =1)
            
        scale = np.nanmean((self.X[:,:,1:self.lookback] - self.X[:,:,0:(self.lookback-1)])**2, axis = -1)
        if pred is not None:
            while len(pred.shape) < len(self.B.shape) :
                pred = np.expand_dims(pred,1)           
            
            pred = np.concatenate([ np.nan_to_num(self.B,nan = 0),pred],1)
        else:
            pred = self.B
        
        return metrics(pred, Y, scale, mask)
    
    def add_Base_forecasts(self, model_pred,forecaster):     
        
         
       
        if forecaster in self.forecasters:         
            position = self.forecasters.index(forecaster)
            
            self.B[:,position,::] = model_pred

        else:
            self.forecasters += [forecaster]
            model_pred =  np.expand_dims(model_pred, 1)
            if self.B is None:
                self.B = model_pred
            else:
                self.B = np.concatenate([self.B, model_pred], 1)
            
            
    def XZ_globalReg(self, L, rolling_step, mode = 'Tree', rescale = False):       

        if L + self.horizon  > self.lookback:
            raise ValueError("The window_size + forecasting "
                                 "horizon would exceed the length of the "
                                 "given timeseries!")
        Zy_train = []
        Zy_test = []
        mask = self.mask
        weight = self.weight
        X_test = self.X[:,:,(self.lookback - L): ]
        mask_test = mask[:,self.lookback: ]
        weight_test = weight[:,self.lookback: ]
        
        rollings = (self.lookback - L - self.horizon)//rolling_step
        
        X_train = [self.X[:,:,(self.lookback - L - self.horizon - i*rolling_step):(self.lookback - self.horizon - i*rolling_step)] for i in range(rollings) ]       
        Y_train = [self.X[:,:,(self.lookback  - self.horizon - i*rolling_step):(self.lookback - i*rolling_step)] for i in range(rollings) ] 
        mask_train = [mask[:,(self.lookback  - self.horizon - i*rolling_step):(self.lookback - i*rolling_step)] for i in range(rollings) ]  
        weight_train = [weight[:,(self.lookback  - self.horizon - i*rolling_step):(self.lookback - i*rolling_step)] for i in range(rollings) ]  
        
        idx_tr = [self.df_idx.values  for i in range(rollings) ]
        tdx_tr = [self.Xt_idx[(self.lookback  - self.horizon - i*rolling_step):(self.lookback - i*rolling_step)] for i in range(rollings) ] 
        
        idx_ts = [self.df_idx.values]
        tdx_ts = [self.Yt_tidx]

               
           
        if self.Z is not None:
            Z_test  =  self.Z[:,:,(self.lookback - L):]
            Zx_test = Z_test[:,:,0:L]
            Zy_test = Z_test[:,:,L:L+self.horizon]         
                        
            Z_train = [self.Z[:,:,(self.lookback - L - self.horizon - i*rolling_step):(self.lookback - i*rolling_step)] for i in range(rollings) ]

                
        if mode == 'seq2seq':         
            X_test = [X_test]
            if rescale:
                X_train, scale_y_tr = self.rescale_y(X_train)
                X_test, scale_y_ts = self.rescale_y(X_test)
            else:
                scale_y_tr = [None for i in range(rollings)]
                scale_y_ts = [None]                  
            
            
            
            if len(self.unknown_features)>2:
                unknown_features_tr = [ np.concatenate([X_train[i],Z_train[i][:,self.unknown_features_idx[1:],:L]],1) for i in range(rollings)]
                unknown_features_ts = np.concatenate([X_test[0], Z_test[:,self.unknown_features_idx[1:],:L]],1)                
            elif len(self.unknown_features) == 2 :                            
                unknown_features_tr = [ np.concatenate([X_train[i],Z_train[i][:,:,:L]],1) for i in range(rollings)]
                unknown_features_ts = np.concatenate([X_test[0], Z_test[:,:,:L]],1)
            else:
                unknown_features_tr = X_train 
                unknown_features_ts = X_test[0]
                
            
            target_features = self.dynamic_known_features_num_idx + self.dynamic_known_features_cat_idx
            if len(target_features)  >1:
                 
                target_features_tr = [Z_train[i][:,target_features,:] for i in range(rollings)]
                target_features_ts = Z_test[:,target_features,:]
            elif len(target_features) == 1:
                 
                target_features_tr = [Z_train[i][:,target_features,:] for i in range(rollings)]
                target_features_ts = Z_test[:,target_features,:]
            else:
                target_features_tr = [None for i in range(rollings)]
                target_features_ts = None           
            
            
            static_features = self.static_known_features_idx 
            if len(static_features)  >1:
                 
                static_features_tr = [Z_train[i][:,static_features,0] for i in range(rollings)]
                static_features_ts = Z_test[:,static_features,0]
                
            elif len(static_features) == 1:
                 
                static_features_tr = [Z_train[i][:,:,0] for i in range(rollings)]
                static_features_ts = Z_test[:,:,0]
            else:
                static_features_tr = [None for i in range(rollings)]
                static_features_ts = None             
            
            data_tr = HGT(
                  edge_index = [self.edge_index for i in range(rollings)],
                  edge_weight = [self.edge_weight for i in range(rollings)],
                  
                  unknown_features = unknown_features_tr,
                  static_features = static_features_tr,
                  targets = Y_train,
                  targets_mask = [np.expand_dims(mask_train[i],1) for i in range(rollings)],
                  target_features = target_features_tr,
                  scales = scale_y_tr,
                  tdx = tdx_tr,
                  idx = idx_tr,
                  )
            
            data_ts = HGT(
                  edge_index = [self.edge_index],
                  edge_weight = [self.edge_weight],
                  
                  unknown_features = [unknown_features_ts],
                  static_features = [static_features_ts],
                  targets = [np.expand_dims(self.Y,1)],
                  targets_mask = [np.expand_dims(mask_test,1)],
                  target_features = [target_features_ts],
                  scales = scale_y_ts,
                  tdx = tdx_ts,
                  idx = idx_ts
                  )
            return data_tr, data_ts
            
            '''    
            return ({'X_tr':X_train,
                    'Y_tr': Y_train,
                    'Z_tr': Z_train,
                    'mask_tr': mask_train, 
                    'weight_tr': weight_train, 
                    'idx_tr':idx_tr,
                    'tidx_tr':tidx_tr,
                    'scale_y_tr': scale_y_tr,
                    },
                    {'X_ts': X_test,
                    'Z_ts': Z_test, 
                    'mask_ts': mask_test,
                    'weight_ts': weight_test, 
                    'idx_ts':idx_ts,
                    'tidx_ts':tdx_ts,
                    'scale_y_ts': scale_y_ts,
                    },
                    {'features_num_idx': self.features_num_idx,
                    'features_cat_idx': self.features_cat_idx,
                    'static_known_features_idx': self.static_known_features_idx,
                    'dynamic_known_features_num_idx': self.dynamic_known_features_num_idx,
                    'dynamic_known_features_cat_idx': self.dynamic_known_features_cat_idx,
                    },
                    {'edge_index' : self.edge_index,
                    'edge_weight': self.edge_weight,                    
                    },
                    )     
            '''      
        elif mode == 'Tree' :
            X_train = np.concatenate(X_train,0) 
            Y_train = np.concatenate(Y_train,0)        
            mask_train = np.concatenate(mask_train,0) 
            weight_train = np.concatenate(weight_train,0) 
            
            if self.Z is not None:
                Z_train = np.concatenate(Z_train,0)
                Zx_train = Z_train[:,:,0:L]
                Zy_train = Z_train[:,:,L:L+self.horizon]
                X_test = np.concatenate([X_test,Zx_test],1)  
                X_train = np.concatenate([X_train,Zx_train],1)
            
            ntr, _ , _ = X_train.shape
            nts, _ , _ = X_test.shape            
            
            Xtr,Xts,Ztr,Zts = [ ],[ ],[ ],[ ]  ## position 0 is the lags of target
            
            for i, f in zip(range(len(self.feature_names)),self.feature_names):
                if f in self.static_known_features:
                    Xtr += [ X_train[:,i,0:1] ]
                    Xts += [ X_test[:,i,0:1] ]
                else:
                    if f in self.dynamic_known_features_num + self.dynamic_known_features_cat:  
                        #Xtr += [ X_train[:,i,:],Zy_train[:,i-1,:] ]
                       # Xts += [ X_test[:,i,:],Zy_test[:,i-1,:] ]
                        Xtr += [ X_train[:,i,:]]
                        Xts += [ X_test[:,i,:]]
                        Ztr += [Zy_train[:,i-1,:]]
                        Zts +=[Zy_test[:,i-1,:]]
                    else:
                        Xtr += [ X_train[:,i,:]]
                        Xts += [ X_test[:,i,:] ]
                                            

            
            return {
                'X_tr': np.concatenate( Xtr,-1),
                'Z_tr': np.concatenate( Ztr,-1) if len(Ztr)>0 else Ztr,
                'Y_tr': Y_train.reshape(ntr,-1),
                'mask_tr': mask_train.reshape(ntr,-1),
                'weight_train': weight_train.reshape(ntr,-1),
                'X_ts': np.concatenate( Xts,-1),
                'Z_ts': np.concatenate( Zts,-1)  if len(Zts)>0 else Zts,
                'mask_ts': mask_test,
                'weight_ts': weight_test,
            }
                
    def reduce_mem_usage(self):
        self.X = self.X.astype(np.float32)
        self.B = self.B.astype(np.float32) if self.B is not None else self.B 
        self.Y = self.Y.astype(np.float32) if self.Y is not None else self.Y 
        self.Z = self.Z.astype(np.float32) if self.Z is not None else self.Z    
        self.MetaFeat = self.MetaFeat.astype(np.float32)  if self.MetaFeat is not None else self.MetaFeat
        self.mask = self.mask.astype(np.int8) if self.mask is not None else self.mask
        self.weight = self.weight.astype(np.float16) if self.weight is not None else self.weight
            
    @staticmethod
    def rescale_y(y):
        ## only for postive forecasting
        eps = 1e-8
        scales = []
        y_scale = y.copy()
        for i in range(len(y)):
            
            scale_sum = np.nansum(y[i]*(y[i]>0),axis= -1,keepdims = True) ## avoid sales of -1
            scale_count = np.nansum((y[i]>0),axis= -1, keepdims = True)    
            scale = scale_sum /(scale_count + eps) 
            y_scale[i] = y_scale[i]/(scale + eps)

            scales.append(scale)

        return y_scale, scales          
            
        
class RawData():
    def __init__(self, 
                 npdata: pd.DataFrame,
                 df_idx: pd.MultiIndex, 
                 time_idx: pd.Index,
                 features_num: List[str],
                 features_cat: List[str],
                 static_known_features : List[str]= None,
                 dynamic_known_features_num : List[str]= None,
                 dynamic_known_features_cat : List[str]= None,
                 mask: np.ndarray = None,
                 weight: np.ndarray = None,
                 edge_index:List[Union[np.ndarray, None]] = None,
                 edge_weight: List[Union[np.ndarray, None]] = None,
                 ):

        self.df_idx = df_idx
        self.time_idx = time_idx
        self.features_num = features_num  ### all the num features
        self.features_cat = features_cat  ### all the cat features
        self.npdata = npdata
        self.N = len(df_idx)
        self.C = len(features_num) +  len(features_cat)
        self.T = len(time_idx)
        self.mask = mask
        self.weight = weight
        self.static_known_features = static_known_features  ## static features are all known category features, should be a subset of features_cat and known_features_cat
        self.dynamic_known_features_num =  dynamic_known_features_num  ## a subset of features_num       
        self.dynamic_known_features_cat =  dynamic_known_features_cat  ## a subset of features_cat, including static features  
        self.edge_index = edge_index
        self.edge_weight = edge_weight
        
    def data_segment(
            self, 
            W : int, 
            H :int, 
            nseg :int = 10, 
            step: int = None):
        """
        :param data: shape(N,C,T) time data:shape(N,T)
        :return: s个shot样本 shape为(s,N,C,W+H) s个时间样本 shape 为(s,N,W+H)
        """
        if W + H > self.T:
            raise ValueError("The window_size + forecasting "
                                 "horizon would exceed the length of the "
                                 "given timeseries!")
            
        window_end = self.T
        
        if step is None:
            step = H
        data_slots = []
        
        n = 0
        while n < nseg:
            if (window_end - W - H) < 0 :
                break
            data_slots.append(
                RawData(
                npdata = self.npdata[:,:,(window_end - W - H):window_end],        
                df_idx = self.df_idx,
                time_idx = self.time_idx[(window_end - W - H):window_end],
                features_num = self.features_num,
                features_cat = self.features_cat,                
                static_known_features = self.static_known_features,
                dynamic_known_features_num = self.dynamic_known_features_num,
                dynamic_known_features_cat = self.dynamic_known_features_cat,
                mask = self.mask[:,(window_end - W - H):window_end] if self.mask is not None else None,
                weight = self.weight[:,(window_end - W - H):window_end] if self.weight is not None else None,
                edge_index = self.edge_index,
                edge_weight = self.edge_weight,
                )
                )
            window_end -= step
            n += 1
            
        return data_slots
    
    
    def Raw2Meta(self,H):
        """
        H : max forecasting horizon

        """    
        return MetaDataset(        
                X = np.expand_dims(self.npdata[:,0,0:(self.T-H)],1), 
                Y = self.npdata[:,0,(self.T-H):], 
                Z = self.npdata[:,1:,:], 
                Xt_idx = self.time_idx[0:(self.T-H)],
                Yt_tidx = self.time_idx[(self.T-H):], 
                df_idx = self.df_idx,
                features_num = self.features_num,
                features_cat = self.features_cat,
                static_known_features = self.static_known_features,
                dynamic_known_features_num = self.dynamic_known_features_num,
                dynamic_known_features_cat = self.dynamic_known_features_cat,
                mask = self.mask,    
                weight = self.weight,  
                edge_index = self.edge_index,
                edge_weight = self.edge_weight,
                
                )
        
        

       
class Md_utils:

    def Df_to_rawdata(
            df : pd.DataFrame,
            idx : List[str], 
            tdx : str,
            target: List[str], 
            features_num: List[str], 
            features_cat: List[str],
            static_known_features : List[str] = None ,
            dynamic_known_features_num :List[str] =None, 
            dynamic_known_features_cat :List[str] =None,             
            mask_idx : str = None,
            weight : pd.DataFrame = None,
            edge_index:List[Union[np.ndarray, None]] = None,
            edge_weight: List[Union[np.ndarray, None]] = None,
            ):
        
        df = df.replace([np.inf, -np.inf], np.nan)
        df = df.groupby(idx, sort=False).apply(lambda x: x.ffill().bfill())    
        
        x = df.pivot(index = idx,columns=tdx)
        df_idx = x.index
        
        features = target + features_num + features_cat
        time_idx =  x[features[0]].columns
        
        if mask_idx is not None:
            mask = x[mask_idx].applymap(lambda x: 0 if np.isnan(x) else 1 ).values
        else:
            mask = x[features[0]].applymap(lambda x: 0 if np.isnan(x) else 1 ).values
        x =  [np.expand_dims( x[features[i]].values, 1) for i in range(len(features))]
        
        if weight is not None:
            weight = weight.pivot(index = idx,columns=tdx).values
        else:
            weight = mask
        
        return RawData(
                npdata = np.concatenate(x,1), 
                df_idx = df_idx,
                time_idx = time_idx ,
                features_num = features_num, 
                features_cat = features_cat, 
                static_known_features = static_known_features, 
                dynamic_known_features_num = dynamic_known_features_num,
                dynamic_known_features_cat = dynamic_known_features_cat,
                mask = mask, 
                weight = weight,
                edge_index = edge_index,
                edge_weight = edge_weight,
                )
            
            
            
    def md_concat(md_list,fidx = None, transform = None):
        N = len(md_list)
        fidx = [] if fidx is None else fidx    
        return MetaDataset(        
            X = np.concatenate([md_list[i].X for i in range(N)],0), 
            B = np.concatenate([md_list[i].B[:,fidx,:] for i in range(N)],0) if md_list[0].B is not None else None , 
            Y = np.concatenate([md_list[i].Y for i in range(N)],0) if md_list[0].Y is not None else None, 
            Z = np.concatenate([md_list[i].Z for i in range(N)],0) if md_list[0].Z is not None else None, 
            MetaFeat = np.concatenate([md_list[i].MetaFeat for i in range(N)],0) if md_list[0].MetaFeat is not None else None, 
            transform = transform,
            Xt_idx = None,
            Yt_tidx = None,
            df_idx = None,
            features_num = md_list[0].features_num,
            features_cat = md_list[0].features_cat,
            forecasters = [md_list[0].forecasters[f] for f in fidx ],
            static_known_features = md_list[0].static_known_features,
            dynamic_known_features_num = md_list[0].dynamic_known_features_num,
            dynamic_known_features_cat = md_list[0].dynamic_known_features_cat,
            mask = np.concatenate([md_list[i].mask for i in range(N)],0) if md_list[0].mask is not None else None,  
            weight = np.concatenate([md_list[i].weight for i in range(N)],0) if md_list[0].weight is not None else None, 
            )
        
        
   
    def metadataclass_update(df_slots):
            
            for i in range(len(df_slots)):
                   
                    s = MetaDataset(        
                            X = df_slots[i].X,
                            B = df_slots[i].B,
                            Y = df_slots[i].Y,
                            Z = df_slots[i].Z,
                            MetaFeat = df_slots[i].MetaFeat,
                            transform = df_slots[i].transform,
                            Xt_idx = df_slots[i].Xt_idx,
                            Yt_tidx = df_slots[i].Yt_tidx,
                            df_idx = df_slots[i].df_idx,
                            features_num = df_slots[i].features_num,
                            features_cat = df_slots[i].features_cat,
                            forecasters = df_slots[i].forecasters,
                            static_known_features = df_slots[i].static_known_features,
                            dynamic_known_features_num = df_slots[i].dynamic_known_features_num,
                            dynamic_known_features_cat = df_slots[i].dynamic_known_features_cat,                            
                            mask = df_slots[i].mask ,
                            weight = df_slots[i].weight,
                            edge_index = df_slots[i].edge_index,
                            edge_weight = df_slots[i].edge_weight,
                                            
                            )
                    df_slots[i] = s
            return df_slots




# -*- coding: utf-8 -*-
"""
Created on Tue Dec  6 15:54:28 2022

@author: MSH
"""
import numpy as np

def metrics (pred, y, scale,mask=None):
    
    err = pred - y
    scale = np.where(scale == 0,(err**2).mean(-1), scale)  
    scale = np.where(scale == 0,1, scale)
    if mask is not None:
        err = err * mask
    
    MAE =  np.abs(err).mean(-1).mean(0)
    MSE =  ( err**2 ).mean(-1).mean(0)
    sMAPE1 =  (np.abs(err).sum(-1)/( pred + y+ 1e-6).sum(-1)  ).mean(0)
    sMAPE2 =  (np.abs(err)/( np.abs(pred) + y + 1e-6) ).mean(-1).mean(0)
      
    RMSSE =  np.sqrt((err**2).mean(-1)/ scale).mean(0)
    
    return {'MAE':MAE, 'MSE':MSE, 'sMAPE1' : sMAPE1, 'sMAPE2' : sMAPE2,'RMSSE' : RMSSE}





# -*- coding: utf-8 -*-
"""
Created on Sat Dec  3 20:29:02 2022

@author: MSH
"""

"""Module dealing with data."""


from .load import  MetaDataset, RawData, Md_utils
from .hetero_static_graph_temporal_signal import HeteroStaticGraphTemporalSignal, temporal_signal_split,HSGT_concat
__all__ = [

    "MetaDataset",
    "RawData",
    "Md_utils",
    "HeteroStaticGraphTemporalSignal",
    "temporal_signal_split",
    "HSGT_concat",
]



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 22 21:33:17 2022

@author: MSH
"""

from .layers import (
    squeeze_excite,
    FExtractor_cnn,
    FExtractor_cnn_SE,
)


__all__ = [
    "squeeze_excite",
    "FExtractor_cnn",
    "FExtractor_cnn_SE",
    ]



# -*- coding: utf-8 -*-
"""
Created on Mon Dec  5 23:23:42 2022

@author: MSH
"""
import torch
class meta_learning_run():
    
    def __init__(
        self, 
        model, 
        loss_fun,
        optimizer,
        train_dataloader, 
        val_dataloader,
        device,
        epochs,
        scale_loss=False):
        
        self.model = model 
        self.loss_fun = loss_fun 
        self.optimizer = optimizer
        self.train_dataloader = train_dataloader 
        self.val_dataloader = val_dataloader
        self.device = device
        self.scale_loss = scale_loss
        
        
        for t in range(epochs):
            print(f"Epoch {t+1}\n-------------------------------")
            self.train_loop()  
            if self.val_dataloader is not None:
                self.val_loop() 
        print("Done!")
        



    def train_loop(self):
        
        #size = len(self.train_dataloader.dataset)
        total_loss = 0
        for batch, meta in enumerate(self.train_dataloader):
             
            # Compute prediction and loss
            pred = self.model(meta)
            if self.model.type == 'comb':
                y = meta['Y'].to(self.device) 
                mask = meta['mask'].to(self.device)
                if self.scale_loss:
                    lookback = meta['X'].shape[2]
                    scale = torch.nanmean((meta['X'][:,0,1:lookback] - meta['X'][:,0,0:(lookback-1)])**2, axis = -1)
                    scale = torch.where(scale == 0,scale+1, scale)   
                    scale = torch.sqrt(scale).to(self.device)  
                    loss = self.loss_fun(pred * mask/scale.unsqueeze(1), y*mask/scale.unsqueeze(1))                  
                else:
                    loss = self.loss_fun(pred * mask, y*mask)
                
            if self.model.type == 'select':
                y = meta['Yc'].to(self.device) 
                mask = (torch.min(meta['mask'],-1,keepdim=True)[0]>0).to(self.device)
                loss = self.loss_fun(pred * mask, y*mask)
            if self.model.type == 'loss':
                y = meta['Ye'].to(self.device) 
                mask = (torch.min(meta['mask'],-1,keepdim=True)[0]>0).to(self.device)
                loss = self.loss_fun(pred * mask, y*mask)
                    
            
            # Backpropagation
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            total_loss += loss.item()
        total_loss /= batch + 1
        print(total_loss)
        
    
    def val_loop(self):
        #size = len(self.test_dataloader.dataset)
        total_loss = 0
        for batch, meta in enumerate(self.val_dataloader):
            
            pred = self.model(meta)   
            
            if self.model.type == 'comb':
                y = meta['Y'].to(self.device) 
                mask = meta['mask'].to(self.device)
                
                if self.scale_loss:
                    lookback = meta['X'].shape[2]
                    scale = torch.nanmean((meta['X'][:,0,1:lookback] - meta['X'][:,0,0:(lookback-1)])**2, axis = -1)
                    scale = torch.where(scale == 0,scale+1, scale) 
                    scale = torch.sqrt(scale).to(self.device)   
                    loss = self.loss_fun(pred * mask/scale.unsqueeze(1), y*mask/scale.unsqueeze(1))                  
                else:
                    loss = self.loss_fun(pred * mask, y*mask)
                
            if self.model.type == 'select':
                y = meta['Yc'].to(self.device) 
                mask = (torch.min(meta['mask'],-1,keepdim=True)[0]>0).to(self.device)
                loss = self.loss_fun(pred * mask, y*mask)
                
            if self.model.type == 'loss':
                y = meta['Ye'].to(self.device) 
                mask = (torch.min(meta['mask'],-1,keepdim=True)[0]>0).to(self.device)
                loss = self.loss_fun(pred * mask, y*mask)              
                      
               
            total_loss += loss.item()
        total_loss /= batch + 1
        print(total_loss)


        
    def forecast(self,test_dataloader):
        #size = len(self.test_dataloader.dataset)
        total_loss = 0
        total_pred = []
        for batch, meta in enumerate(test_dataloader):
            
               
            y = meta['Y'].to(self.device)         
            mask = meta['mask'].to(self.device) 
                
            pred = self.model(meta)          
       
            
            if self.model.type == 'select':
                b = meta['B'].to(self.device)
                best_idx = torch.argmax(pred, 1, keepdim=True)
                one_hot = torch.FloatTensor(pred.shape).to(self.device)
                one_hot.zero_()
                one_hot.scatter_(1, best_idx, 1)
                pred = (one_hot.unsqueeze(2)*b).sum(-2)
                
                #pred = (pred.unsqueeze(2)*b).sum(-2)

            if self.model.type == 'loss':
                b = meta['B'].to(self.device)
                weights = torch.nn.functional.softmax(torch.exp(-pred),dim=-1)
                pred = (weights.unsqueeze(2)*b).sum(-2)
                
            total_pred.append(pred.detach() )
            if len(y)>0:
                loss = self.acc(pred*mask, y*mask)         
                total_loss += loss.item()   
                
        total_loss /= batch + 1
        print(total_loss)               
        
        return total_loss, torch.cat(total_pred,0)
    
    def acc(self,pred, y):
        
        return torch.nn.functional.mse_loss(pred, y)
        
        
        
        
        
        
        
        



# -*- coding: utf-8 -*-
"""
Created on Mon Dec  5 17:12:32 2022

@author: MSH
"""


import torch


from .layers import (
    FExtractor_cnn,
    FExtractor_cnn_SE,
    )

class MetaComb_se(torch.nn.Module):


    def __init__(
            self,
            n_features,
            n_forecasters,
            window_x, 
            horizon, 
            Use_z = True, 
            n_known_features = None,
            x_filters = 64,
            z_filters = 64,
            p=0.5 ,
            device = 'cuda',
            ):
        
        super().__init__()
        self.type = 'comb'
        self.n_features = n_features
        self.n_forecasters = n_forecasters  
        self.window_x = window_x     
        self.horizon = horizon          
        self.dim_extract_features = x_filters + z_filters  if Use_z else x_filters       
   
        self.device = device
        self.n_known_features = n_features if n_known_features is None else n_known_features
        self.Use_z =  Use_z       
        
        self.norm_layer_x = torch.nn.InstanceNorm1d(n_features+1)
        
        self.norm_layer_z = torch.nn.InstanceNorm1d(n_known_features)
        
        self.fcn_y = FExtractor_cnn_SE(n_features+1, x_filters, window_x, p=p)
        
        if self.n_known_features>0:
            self.fcn_z = FExtractor_cnn_SE(self.n_known_features, z_filters, horizon, p=p)
        
        self.weight = torch.nn.Sequential(
            torch.nn.Linear(self.dim_extract_features,n_forecasters),                         
            torch.nn.Softmax(dim = 1),
            )
        
        self.LossPred = torch.nn.Sequential(
                torch.nn.Linear(self.dim_extract_features,n_forecasters),
                torch.nn.ReLU(),
                )
        
        
    def forward(self, meta):
       
        x = meta['X'].to(self.device)
        b = meta['B'].to(self.device)
        z = meta['Z'].to(self.device)       
        n_samples, _ ,_  = x.shape
                
        x = self.norm_layer_x(x)
        x = self.fcn_y(x)
        
        if self.Use_z:        
            z = self.norm_layer_z(z)
            z = self.fcn_z(z)        
            xz = torch.concat([x,z],dim = -1)
            weights = self.weight(xz.view(n_samples,-1))
        else:
            weights = self.weight(x.view(n_samples,-1))
        
        y = weights.unsqueeze(-1) * b   

        return y.sum(-2)


class MetaComb(torch.nn.Module):


    def __init__(
            self,
            n_features,
            n_forecasters,
            window_x, 
            horizon, 
            Use_z = True, 
            n_known_features = None,
            x_filters = 64,
            z_filters = 64,
            p=0.5 ,
            device = 'cuda',
            ):
        
        super().__init__()
        self.type = 'comb'
        self.n_features = n_features
        self.n_forecasters = n_forecasters  
        self.window_x = window_x     
        self.horizon = horizon          
        self.dim_extract_features = x_filters + z_filters  if Use_z else x_filters       
   
        self.device = device
        self.n_known_features = n_features if n_known_features is None else n_known_features
        self.Use_z =  Use_z       
        
        self.norm_layer_x = torch.nn.InstanceNorm1d(n_features+1)
        
        self.norm_layer_z = torch.nn.InstanceNorm1d(n_known_features)
        
        self.fcn_y = FExtractor_cnn(n_features+1, x_filters, window_x, p=p)
        
        if self.n_known_features>0:
            self.fcn_z = FExtractor_cnn(self.n_known_features, z_filters, horizon, p=p)
        
        self.weight = torch.nn.Sequential(
            torch.nn.Linear(self.dim_extract_features,n_forecasters),                         
            torch.nn.Softmax(dim = 1),
            )
        
        self.LossPred = torch.nn.Sequential(
                torch.nn.Linear(self.dim_extract_features,n_forecasters),
                torch.nn.ReLU(),
                )
        
        
    def forward(self, meta):
       
        x = meta['X'].to(self.device)
        b = meta['B'].to(self.device)
        z = meta['Z'].to(self.device)       
        n_samples, _ ,_  = x.shape
                
        x = self.norm_layer_x(x)
        x = self.fcn_y(x)
        
        if self.Use_z:        
            z = self.norm_layer_z(z)
            z = self.fcn_z(z)        
            xz = torch.concat([x,z],dim = -1)
            weights = self.weight(xz.view(n_samples,-1))
        else:
            weights = self.weight(x.view(n_samples,-1))
        
        y = weights.unsqueeze(-1) * b   

        return y.sum(-2)  
    

class MetaSelection(MetaComb):   
    
    def __init__(self, 
                 n_features,
                 n_forecasters, 
                 window_x, 
                 horizon, 
                 Use_z = True, 
                 n_known_features = None, 
                 x_filters = 64,
                 z_filters = 64,
                 p=0.5 ,
                 device = 'cuda'):        
       
        super().__init__(
                n_features = n_features, 
                n_forecasters = n_forecasters,
                window_x = window_x, 
                horizon = horizon, 
                Use_z = Use_z,
                n_known_features = n_known_features,
                x_filters = x_filters,
                z_filters = z_filters,
                p = p ,
                device  = device,
                )
        self.type = 'select'    
        
    def forward(self, meta, Use_z=True):
        x = meta['X'].to(self.device)
        b = meta['B'].to(self.device)
        z = meta['Z'].to(self.device)       
        n_samples, _ ,_  = x.shape
        # Normalize
        
        x = self.norm_layer_x(x)
        x = self.fcn_y(x)
        
        
        if self.Use_z:        
            z = self.norm_layer_z(z)
            z = self.fcn_z(z)        
            xz = torch.concat([x,z],dim = -1)
            weights = self.weight(xz.view(n_samples,-1))
        else:
            weights = self.weight(x.view(n_samples,-1))     

        return weights

class MetaLoss(MetaComb):
    
    def __init__(self, 
                 n_features,
                 n_forecasters, 
                 window_x, 
                 horizon, 
                 Use_z = True, 
                 n_known_features = None, 
                 x_filters = 64,
                 z_filters = 64,
                 p=0.5 ,
                 device = 'cuda'):        
       
        super().__init__(
                n_features = n_features, 
                n_forecasters = n_forecasters,
                window_x = window_x, 
                horizon = horizon, 
                Use_z = Use_z,
                n_known_features = n_known_features,
                x_filters = x_filters,
                z_filters = z_filters,
                p = p ,
                device  = device,
                )
        
        self.type = 'loss'        
        
    def forward(self, meta, Use_z=True):
        x = meta['X'].to(self.device)
        b = meta['B'].to(self.device)
        z = meta['Z'].to(self.device)       
        n_samples, _ ,_  = x.shape
        # Normalize
        
        x = self.norm_layer_x(x)
        x = self.fcn_y(x)
                
        if self.Use_z:        
            z = self.norm_layer_z(z)
            z = self.fcn_z(z)        
            xz = torch.concat([x,z],dim = -1)
            loss = self.LossPred(xz.view(n_samples,-1))
        else:               
            loss = self.LossPred(x.view(n_samples,-1))         

        return loss
    
from deepdow.layers.misc import CovarianceMatrix
from deepdow.layers import SparsemaxAllocator,SoftmaxAllocator,AnalyticalMarkowitz, NumericalRiskBudgeting

class MetaComb_risk(MetaComb):
    
    def __init__(self, 
                 n_features,
                 n_forecasters, 
                 window_x, 
                 horizon, 
                 Use_z = True, 
                 n_known_features = None, 
                 x_filters = 64,
                 z_filters = 64,
                 p=0.5 ,
                 device = 'cuda'):        
       
        super().__init__(
                n_features = n_features, 
                n_forecasters = n_forecasters,
                window_x = window_x, 
                horizon = horizon, 
                Use_z = Use_z,
                n_known_features = n_known_features,
                x_filters = x_filters,
                z_filters = z_filters,
                p = p ,
                device  = device,
                )
        
        self.type = 'comb'
        self.weight = torch.nn.Sequential(
            torch.nn.Linear(self.dim_extract_features,n_forecasters),                         
            torch.nn.Softmax(dim = 1),
            )   
        self.cov_layer = CovarianceMatrix(
            sqrt=True, shrinkage_strategy="scaled_identity"
        )     
        
        self.allocater = NumericalRiskBudgeting(n_assets = n_forecasters)
        
    def forward(self, meta, Use_z=True):
        x = meta['X'].to(self.device)
        b = meta['B'].to(self.device)
        z = meta['Z'].to(self.device)     
        e = meta['E'].to(self.device) 
        n_samples, _ ,_  = x.shape
        covmat_sqrt = self.cov_layer(e.swapaxes(1,-1))        
        x = self.norm_layer_x(x)
        x = self.fcn_y(x)
        
        if self.Use_z:        
            z = self.norm_layer_z(z)
            z = self.fcn_z(z)        
            xz = torch.concat([x,z],dim = -1)
            weights = self.weight(xz.view(n_samples,-1))
        else:
            weights = self.weight(x.view(n_samples,-1))
        
        weights = self.allocater(covmat_sqrt,weights)
        y = weights.unsqueeze(-1) * b   

        return y.sum(-2)


import torch
import numpy as np
from typing import List, Union,Tuple
import pandas as pd

import torch.utils.data
from torch.utils.data.dataloader import default_collate
from torch_geometric.data import Data, HeteroData, Dataset, Batch
from collections.abc import Mapping, Sequence
from torch.utils.data.sampler import Sampler
from torch.utils.data import DataLoader,Dataset

Edge_Index = List[Union[np.ndarray, None]]
Edge_Weight = List[Union[np.ndarray, None]]
NNets = Union[int, None]
Node_Features_Timevarying = List[Union[np.ndarray, None]]
Node_Features_Static = List[Union[np.ndarray, None]]
Targets = List[Union[np.ndarray, None]]
Target_features = List[Union[np.ndarray, None]]
Additional_Features = List[np.ndarray]


                        
class HeteroStaticGraphTemporalSignal(object):
    r"""A data iterator object to contain a static graph with a dynamically
    changing constant time difference temporal feature set (multiple signals).
    The node labels (target) are also temporal. The iterator returns a single
    constant time difference temporal snapshot for a time period (e.g. day or week).
    This single temporal snapshot is a Pytorch Geometric Data object. Between two
    temporal snapshots the features and optionally passed attributes might change.
    However, the underlying graph is the same.

    Args:
        edge_index (Numpy array): Index tensor of edges.
        edge_weight (Numpy array): Edge weight tensor.
        features (List of Numpy arrays): List of node feature tensors.
        targets (List of Numpy arrays): List of node label (target) tensors.
        **kwargs (optional List of Numpy arrays): List of additional attributes.
    """

    def __init__(
        self,
        edge_index: Edge_Index,
        edge_weight: Edge_Weight,        
        unknown_features: Node_Features_Timevarying,
        static_features: Node_Features_Static,
        targets: Targets,
        targets_mask: Targets,
        target_features:Target_features,
        scales:List[Union[np.ndarray, None]],
        tdx:List[Union[np.ndarray, None]],
        idx:List[Union[np.ndarray, None]],
        **kwargs: Additional_Features
    ):
        self.edge_index = edge_index
        self.edge_weight = edge_weight        
        self.unknown_features = unknown_features
        self.static_features = static_features
        self.targets = targets
        self.targets_mask = targets_mask
        self.target_features = target_features
        self.additional_feature_keys = []
        self.scales = scales
        self.tdx = tdx
        self.idx = idx
        for key, value in kwargs.items():
            setattr(self, key, value)
            self.additional_feature_keys.append(key)
        self._check_temporal_consistency()
        self._set_snapshot_count()

    def _check_temporal_consistency(self):
        assert len(self.unknown_features) == len(
            self.targets
        ), "Temporal dimension inconsistency."
        for key in self.additional_feature_keys:
            assert len(self.targets) == len(
                getattr(self, key)
            ), "Temporal dimension inconsistency."

    def _set_snapshot_count(self):
        self.snapshot_count = len(self.unknown_features)

    def _get_edge_index(self, time_index: int):
        if self.edge_index[time_index] is None:
            return self.edge_index[time_index]
        else:
            return torch.LongTensor(self.edge_index[time_index])

    def _get_edge_weight(self, time_index: int):
        if self.edge_weight[time_index] is None:
            return self.edge_weight[time_index]
        else:
            return torch.FloatTensor(self.edge_weight[time_index])

    def _get_features(self, time_index: int):
        if self.unknown_features[time_index] is None:
            return self.unknown_features[time_index]
        else:
            return torch.FloatTensor(
                    np.nan_to_num( self.unknown_features[time_index])
                    )

    def _get_target(self, time_index: int):
        if self.targets[time_index] is None:
            return self.targets[time_index]
        else:
            if self.targets[time_index].dtype.kind == "i":
                return torch.LongTensor(
                        np.nan_to_num( self.targets[time_index])
                        )
            elif self.targets[time_index].dtype.kind == "f":
                return torch.FloatTensor(
                        np.nan_to_num( self.targets[time_index])
                        )

    def _get_target_mask(self, time_index: int):
        if self.targets_mask[time_index] is None:
            return self.targets_mask[time_index]
        else:
            if self.targets_mask[time_index].dtype.kind == "i":
                return torch.LongTensor(
                        np.nan_to_num( self.targets_mask[time_index])
                        )
            elif self.targets_mask[time_index].dtype.kind == "f":
                return torch.FloatTensor(
                        np.nan_to_num( self.targets_mask[time_index])
                        )

    def _get_target_features(self, time_index: int):
        if self.target_features[time_index] is None:
            return self.target_features[time_index]
        else:
            return torch.FloatTensor(
                    np.nan_to_num( self.target_features[time_index])
                    )

    def _get_static_features(self, time_index: int):
        if self.static_features[time_index] is None:
            return self.target_features[time_index]
        else:
            return torch.FloatTensor(
                    np.nan_to_num( self.static_features[time_index],nan = 0)
                    )

    def _get_target_scale(self, time_index: int):
        if self.scales[time_index] is None:
            return self.scales[time_index]
        else:
            return torch.FloatTensor(
                    np.nan_to_num( self.scales[time_index])
                    )

    def _get_target_tdx(self, time_index: int):
        if self.tdx[time_index] is None:
            return self.tdx[time_index]
        else:
            return torch.FloatTensor(self.tdx[time_index])

    def _get_additional_feature(self, time_index: int, feature_key: str):
        feature = getattr(self, feature_key)[time_index]
        if feature.dtype.kind == "i":
            return torch.LongTensor(feature)
        elif feature.dtype.kind == "f":
            return torch.FloatTensor(feature)

    def _get_additional_features(self, time_index: int):
        additional_features = {
            key: self._get_additional_feature(time_index, key)
            for key in self.additional_feature_keys
        }
        return additional_features

    def __get_item__(self, time_index: int):
        x = self._get_features(time_index)
        edge_index = self._get_edge_index(time_index)
        edge_weight = self._get_edge_weight(time_index)
        y = self._get_target(time_index)
        mask = self._get_target_mask(time_index)
        z = self._get_target_features(time_index)
        scale = self._get_target_scale(time_index)
        static_f = self._get_static_features(time_index)
        
        additional_features = self._get_additional_features(time_index)
        snapshot = HeteroData()
        snapshot['node'].x = x
        snapshot['node'].y = y
        snapshot['node'].mask = mask
        snapshot['node'].z = z
        snapshot['node'].scale = scale
        snapshot['tdx'] = self.tdx[time_index]
        snapshot['idx'] = self.idx[time_index]        
        
        snapshot['nodenet'].edge_index = edge_index        
        snapshot['nodenet'].edge_attr = edge_weight
        
        
        snapshot['node'].f = static_f
 
        return snapshot

    def __next__(self):
        if self.t < len(self.unknown_features):
            snapshot = self.__get_item__(self.t)
            self.t = self.t + 1
            return snapshot
        else:
            self.t = 0
            raise StopIteration

    def __iter__(self):
        self.t = 0
        return self
    
    def __len__(self):
        """Compute length."""
        return len(self.unknown_features)
    
    @staticmethod   
    def _collate_fn(
        batches: List[HeteroData]
        ) -> HeteroData:
        
        return Batch.from_data_list(batches)

    def to_dataloader(
        self, train: bool = True, batch_size: int = 1, batch_sampler: Union[Sampler, str] = None, **kwargs
    ) -> DataLoader:
        
        """
        Get dataloader from dataset.
        """
        data = iter(self)
        
        dataset = [next(data) for j in range(len(self))]
        
        default_kwargs = dict(
            shuffle=train,
            drop_last=train and len(self) > batch_size,
            collate_fn=self._collate_fn,
            batch_size=batch_size,
            batch_sampler=batch_sampler,
        )
        default_kwargs.update(kwargs)
        kwargs = default_kwargs

        return DataLoader(
            dataset,
            **kwargs,
        ) 

 


def temporal_signal_split(
    data_iterator, train_ratio: float = 0.8, H : float = 1
) -> Tuple[HeteroStaticGraphTemporalSignal, HeteroStaticGraphTemporalSignal]:
    r"""Function to split a data iterator according to a fixed ratio.

    Arg types:
        * **data_iterator** *(Signal Iterator)* - Node features.
        * **train_ratio** *(float)* - Graph edge indices.

    Return types:
        * **(train_iterator, test_iterator)** *(tuple of Signal Iterators)* - Train and test data iterators.
    """

    train_snapshots = int(train_ratio * data_iterator.snapshot_count)

    train_iterator = HeteroStaticGraphTemporalSignal(
        data_iterator.edge_index[0:(train_snapshots-H+1)],
        data_iterator.edge_weight[0:(train_snapshots-H+1)],        
        data_iterator.unknown_features[0:(train_snapshots-H+1)],
        data_iterator.static_features[0:(train_snapshots-H+1)],
        
        data_iterator.targets[0:(train_snapshots-H+1)],
        data_iterator.targets_mask[0:(train_snapshots-H+1)],
        data_iterator.target_features[0:(train_snapshots-H+1)],
        
        data_iterator.scales[0:(train_snapshots-H+1)],
        data_iterator.tdx[0:(train_snapshots-H+1)],
        data_iterator.idx[0:(train_snapshots-H+1)],
        
    )

    test_iterator = HeteroStaticGraphTemporalSignal(
        data_iterator.edge_index[train_snapshots:],
        data_iterator.edge_weight[train_snapshots:],
        
        data_iterator.unknown_features[train_snapshots:],
        data_iterator.static_features[train_snapshots:],
        
        data_iterator.targets[train_snapshots:],
        data_iterator.targets_mask[train_snapshots:],
        data_iterator.target_features[train_snapshots:],
        
        data_iterator.scales[train_snapshots:],
        data_iterator.tdx[train_snapshots:],
        data_iterator.idx[train_snapshots:],
        
    )

    return train_iterator, test_iterator



def HSGT_concat(
    data_iterator: List[HeteroStaticGraphTemporalSignal]
) -> HeteroStaticGraphTemporalSignal:
    
    edge_index = []
    edge_weight = []
    
    unknown_features = []
    static_features = []
    targets= []
    targets_mask = []
    target_features = []
    scales =[]
    tdx = []
    idx = []
    
    N =  len(data_iterator)
    
    for i in range(N):
        edge_index += data_iterator[i].edge_index
        edge_weight += data_iterator[i].edge_weight
        unknown_features += data_iterator[i].unknown_features
        static_features += data_iterator[i].static_features
        targets += data_iterator[i].targets
        targets_mask += data_iterator[i].targets_mask
        target_features += data_iterator[i].target_features
        scales += data_iterator[i].scales
        tdx += data_iterator[i].tdx
        idx += data_iterator[i].idx
    
    concated_data = HeteroStaticGraphTemporalSignal(
        edge_index = edge_index,
        edge_weight = edge_weight,
        
        unknown_features = unknown_features,
        static_features =static_features,
        
        targets = targets,
        targets_mask = targets_mask,
        target_features = target_features,
        
        scales = scales,
        tdx = tdx,
        idx = idx,
        
    )
   
    return concated_data





# -*- coding: utf-8 -*-
"""
Created on Thu Dec 22 17:12:33 2022

@author: MSH
"""

import torch
import torch.nn as nn

class squeeze_excite(nn.Module):

    def __init__(
            self, 
            L_in, 
            squeeze_ratio = 0.1,
            ):
        
        super().__init__()
        
        self.squeeze_to = int(L_in * squeeze_ratio)
        
        if self.squeeze_to == 0:
            self.squeeze_to = 1
        
        self.squeeze = nn.Sequential(
                nn.AdaptiveAvgPool1d(self.squeeze_to),
                nn.Linear(self.squeeze_to, L_in//2, bias=False),
                nn.ReLU(),
                nn.Linear(L_in//2, L_in, bias=False),
                nn.Sigmoid(),              
                
                )        
        
    def forward(self, x):
        
        return x * self.squeeze(x)
        

class FExtractor_cnn(torch.nn.Module):

    def __init__(
            self, 
            in_channels, 
            out_channels, 
            window_x,
            kernel_size = 2, 
            stride = 1, 
            dilation = 1, 
            padding = 0,
            p = 0.5,
            ):        
        
        super().__init__()
        self.fcn = nn.Sequential(
            
            nn.Conv1d(
                    in_channels =  in_channels,
                    out_channels = out_channels, 
                    kernel_size= kernel_size if window_x > 16 else 1, 
                    stride= stride, 
                    padding = padding, 
                    dilation = dilation, 
                    groups = 1, 
                    bias = True,
                    ),                    
            nn.Conv1d(
                    in_channels =  out_channels,
                    out_channels = out_channels*2, 
                    kernel_size= kernel_size *2 if window_x > 16 else 1, 
                    stride= stride, 
                    padding = padding, 
                    dilation = dilation, 
                    groups = 1, 
                    bias = True,
                    ),            
            nn.Conv1d(
                    in_channels = out_channels*2,
                    out_channels = out_channels, 
                    kernel_size = kernel_size * 4 if window_x > 16 else 1, 
                    stride = stride, 
                    padding = padding, 
                    dilation = dilation, 
                    groups = 1, 
                    bias = True,
                    ),    
            
            nn.AdaptiveAvgPool1d(1),            
            nn.Dropout(p=p), 
        )
        
        
    def forward(self, x):
        
        return self.fcn(x)
    
    
class FExtractor_cnn_SE(torch.nn.Module):

    def __init__(
            self, 
            in_channels, 
            out_channels, 
            window_x,
            kernel_size = 2, 
            stride = 1, 
            dilation = 1, 
            padding = 0,
            p = 0.5, 
            ):
        
        super().__init__()

        if window_x <= 16:
            kernel_size = 1
            
        
        Lout1 =    ( window_x + 2 * padding - dilation *(kernel_size -1) -1 )/stride + 1
        
        if window_x <= 16:
            Lout2 =    ( Lout1 + 2 * padding - dilation *(kernel_size -1) -1 )/stride + 1
        else:
            Lout2 =    ( Lout1 + 2 * padding - dilation *(kernel_size *2 -1) -1 )/stride + 1
        
        if Lout2 % 1 != 0 :
            raise ValueError("Lout should be integer!")
        
            
        self.fcn = torch.nn.Sequential(
            
            torch.nn.Conv1d(
                    in_channels =  in_channels,
                    out_channels = out_channels, 
                    kernel_size= kernel_size  if window_x > 16 else 1, 
                    stride= stride, 
                    padding = padding, 
                    dilation = dilation, 
                    groups = 1, 
                    bias = True),
            
            squeeze_excite(int(Lout1),0.1),
                    
            torch.nn.Conv1d(
                    in_channels =  out_channels,
                    out_channels = out_channels*2, 
                    kernel_size= kernel_size *2 if window_x > 16 else 1, 
                    stride= stride, 
                    padding=padding, 
                    dilation = dilation, 
                    groups = 1, 
                    bias = True),
            
            squeeze_excite(int(Lout2),0.1),
            
            torch.nn.Conv1d(
                    in_channels = out_channels*2,
                    out_channels = out_channels, 
                    kernel_size = kernel_size * 4 if window_x > 16 else 1, 
                    stride = stride, 
                    padding = padding, 
                    dilation = dilation, 
                    groups = 1, 
                    bias = True),
            
            
            torch.nn.AdaptiveAvgPool1d(1),
            
            torch.nn.Dropout(p=p), 
        )
        
        
    def forward(self, x):
        
        return self.fcn(x)